"""
This module provides classes for defining recommender rules.

Classes:
- RecommenderRule: Represents a recommender rule.
- AllRecommendFunctionsArgs: Represents the arguments for all recommend functions.

"""

from .recommender_rule import AllRecommendFunctionsArgs, RecommenderRule

__all__ = ["AllRecommendFunctionsArgs", "RecommenderRule"]
