"""
This module contains the environments package.

The environments package provides classes and functions for creating and managing game environments.

"""

from typing import Annotated

type HalfMove = Annotated[int, "half move"]
