"""
Module for neural network board evaluators.
"""

from .neural_net_board_eval_args import NeuralNetBoardEvalArgs
from .nn_board_evaluator import NNBoardEvaluator

__all__ = ["NNBoardEvaluator", "NeuralNetBoardEvalArgs"]
